// wuiobjdb.c - implement WUIMAN object database.
// Copyright (c) 1993 by Ron Burk

#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>      // sscanf()

#include "wuidbint.h"
#include "wuierror.h"
#include "wuistd.h"
#include "wuiobjdb.h"


////////////////////////////////////////////////////////////
// TWuiObjects - Provide arrays of AWuiObject pointers

// TWuiObjects - initialize object with desired array size.
TWuiObjects::TWuiObjects(int Size)
    :   TWuiGenericList(Size)
    {   MEMBERASSERT();
    }

// (private)
// Find - Return position of named child

int     TWuiObjects::Find(const TWuiName Which)
    {   MEMBERASSERT();

    for(int iChild = 0; iChild < NObjects(); ++iChild)
        {
        ASSERT(Get(iChild) != NULL);
        if(Get(iChild)->Name() == Which)
            return iChild;
        }
    return -1;
    }

AWuiObject *TWuiObjects::Insert(AWuiObject *NewObject, const TWuiName To)
    {   MEMBERASSERT();
    ASSERT(NewObject != 0);
    return (AWuiObject *)Insert(NewObject, Find(To));
    }

AWuiObject *TWuiObjects::Insert(AWuiObject *NewObject, int Position)
    {   MEMBERASSERT();
    ASSERT(NewObject != 0);
    return (AWuiObject *)TWuiGenericList::Insert((void *)NewObject, Position);
    }

AWuiObject  *TWuiObjects::Get(const TWuiName Which)
    {   MEMBERASSERT();
    int     Position = Find(Which);
    if(Position >= 0)
        return Get(Position);
    else
        return 0;
    }

AWuiObject *TWuiObjects::Remove(const TWuiName Which)
    {   MEMBERASSERT();
    int     Position = Find(Which);
    if(Position >= 0)
        return Remove(Position);
    else
        return 0;
    }

////////////////////////////

//static TWuiRegister RegisterObject(
//    new TWuiObject(C_TOBJECT));

// TWuiObject - An AWuiObject that supports attributes.
TWuiObject::TWuiObject(const TWuiName ObjectName,
                              int NAttributes)
    :   AWuiObject(ObjectName), Attributes_(0)
    {   MEMBERASSERT();
    // note that it is Ok if NAttributes is 0,
    // as the goal here is just to create the array.
    Attributes_ = new TWuiObjects(NAttributes+3);
    // All objects support these methods
    Attributes_->Insert(new TWuiMethod(M_ISA));
    Attributes_->Insert(new TWuiMethod(M_DELETE));
    Attributes_->Insert(new TWuiMethod(M_RENAME));
    // All objects support this property
    ASSERT(P_Persistent != NULL);
    Attributes_->Insert(P_Persistent);
    }


// Set - Handle modifications to "Persistent" property.
//
// 1) making an object persistent makes all its ancestor
//    containers persistent.
// 2) making a container non-persistent makes all its child
//    objects (and their attributes) non-persistent.

long    TWuiObject::Set(TWuiPath Path, TWuiArgs *Args,
                                         const char *Buffer)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("TWuiObject::Set", Path, Args);
    long    Result = AWuiObject::Set(Path, Args, Buffer);
    // if not attribute "p_Persistent", then we don't care
    if(Args->AttributeName() != P_PERSISTENT)
        return AWuiObject::Set(Path, Args, Buffer);
    // if set was not successful, then we don't care
    if(Result != 1 && Result != 0)
        return Result;
    if(Result == 1)  // if enabling persistence
        SetFlags(PERSISTENT);
    // else if we are being made non-persistent
    else if(!Path.Descend())
        {
        ClearFlags(PERSISTENT);
        Path.Ascend();
        TWuiObjects *List = Attributes();
        if(List)
            Broadcast(List, Path, Args, Buffer, 0, TRUE);
        List    = Children();
        if(List)
            Broadcast(List, Path, Args, Buffer, 0, TRUE);
        AWuiObject::Database->DeletePath(Path.AbsolutePath());
        }

    return Result;
    }

TWuiName    TWuiObject::ClassName() const
    {   MEMBERASSERT();
    return C_TOBJECT;
    }
TWuiName    TWuiObject::ParentClassName() const
    {   MEMBERASSERT();
    return C_ROOT;
    }

void    TWuiObject::Copy(const AWuiObject *Other,
                                  int Depth)
    {   MEMBERASSERT();
    AWuiObject::Copy(Other, Depth);
    }

AWuiObject  *TWuiObject::Clone(const TWuiName NewName,
                                  int Depth)
    {   MEMBERASSERT();
    ASSERT(Attributes_ != NULL);
    AWuiObject *Result = new TWuiObject(NewName,
                                  Attributes()->NObjects());
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }
long    TWuiObject::Save(TWuiPath Path, int Depth) const
    {   MEMBERASSERT();
    char    MyClass[TWuiName::MAXNAME+1];
    ClassName().GetName(MyClass);
    char    MyName[TWuiName::MAXNAME+1];
    Name().GetName(MyName);
DEBUG_TRACE(">TWuiObject::Save< name='%s', class='%s'", MyName, MyClass);
    char    MyPath[TWuiPath::MAXPATH];
    strcpy(MyPath, Path.AbsolutePath());
    MyPath[Path.RelativePath()-Path.AbsolutePath()] = '\0';
//    if(MyPath[0] == '0') // if root, do nothing
//        return 0;
    if(!Database->WriteItem(MyPath,
                        MyName, MyClass))
        {
        DEBUG_ERROR("Write of '%s':'%s'='%s' failed",
            Path.AbsolutePath(), MyName, MyClass);
        return WUIMAN_ERR_WRITE_FAILED;
        }
    else
        return AWuiObject::Save(Path, Depth);
    }


TWuiObjects *TWuiObject::Attributes() const
    {   MEMBERASSERT();
    return Attributes_;
    }

int     TWuiObject::IsA(const TWuiName Class,
                                      const TWuiName Method)
    {   MEMBERASSERT();
    if(Class == C_TOBJECT)
        {
        if(Method == "")
            return TRUE;
        else if(GetAttribute(Method) != NULL)
            return TRUE;
        }
    return FALSE;
    }

long    TWuiObject::DefineAttribute(AWuiAttribute *Attr)
    {   MEMBERASSERT();
    if(Attr == NULL)
        {
        DEBUG_ERROR("DefineAttribute(): NULL attribute "
                               "value (assume new failed)");
        return WUIMAN_ERR_NEW_FAILED;
        }
    TWuiObjects *List = Attributes();
    if(List == NULL)
        {
        DEBUG_ERROR("DefineAttribute(): This object does "
            "not support attributes.");
        return WUIMAN_ERR_TARGET_HATES_ATTR;
        }
    if(List->Get(Attr->Name()))
        {
        char    Name[TWuiName::MAXNAME];
        DEBUG_ERROR("DefineAttribute('%s'): Name collision.",
                                 Attr->Name().GetName(Name));
        DumpObject();
        return WUIMAN_ERR_DUPLICATE_NAME;
        }
    List->Insert(Attr);
    return TRUE;
    }


////////////


// TWuiNonContainer - An AWuiObject that supports attributes.
TWuiNonContainer::TWuiNonContainer(const TWuiName ObjectName,
                              int NAttributes)
    :   TWuiObject(ObjectName, NAttributes)
    {   MEMBERASSERT();
    }

// Class constructor -- just return master object.
AWuiObject *TWuiNonContainer::ClassConstructor(
                                        const TWuiName Name)
    {
    return new TWuiNonContainer(Name);
    }



////////////
// 11/06/93

AWuiObject *TWuiContainer::ClassConstructor(const TWuiName Name)
    {
    return new TWuiContainer(Name);
    }

// TWuiContainer - A TWuiObject that supports children.
TWuiContainer::TWuiContainer(const TWuiName ObjectName,
                              int NAttributes, int NChildren)
    :   TWuiObject(ObjectName, NAttributes+1), Children_(0)
    {   MEMBERASSERT();
    // note that it is Ok if NChildren or NAttributes is 0,
    // as the goal here is just to create the arrays.
    Children_ = new TWuiObjects(NChildren);
    }

// Set - Handle modifications to "Persistent" property.
//
// 1) making an object persistent makes all its ancestor
//    containers persistent.
// 2) making a container non-persistent makes all its child
//    objects (and their attributes) non-persistent.

long    TWuiContainer::Set(TWuiPath Path, TWuiArgs *Args,
                                         const char *Buffer)
    {   MEMBERASSERT();
#if 0
    // if not attribute "p_Persistent", then we don't care
    if(Args->AttributeName() != P_PERSISTENT)
        return Result;
    // if not setting or clearing, then we don't care
    if(Buffer == NULL || Buffer[0] == '\0')
        return Result;
    // determine whether enabling or disabling persistence
    int TurnOn = (atoi(Buffer) != 0);

    if(TurnOn)  // if enabling persistence
        {
        if(!Path.Descend()) // if message was for me
            return Result;  // then I am already set
        else    // else, make me persistent too!
            {
            Path.Ascend();
            return AWuiObject::Set(Path, Args, Buffer);
            }
        }
    // else if we are being made non-persistent
    else if(!Path.Descend())
        {
        Path.Ascend();
        TWuiObjects *List = Attributes();
        if(List)
            
        List    = Children();
        if(List)
            ;
        }
    else
#endif
    long    Result = AWuiObject::Set(Path, Args, Buffer);
        return Result;
    }

TWuiName    TWuiContainer::ClassName() const
    {   MEMBERASSERT();
    return C_TCONTAINER;
    }
TWuiName    TWuiContainer::ParentClassName() const
    {   MEMBERASSERT();
    return C_ROOT;
    }

void    TWuiContainer::Copy(const AWuiObject *Other, int Depth)
    {   MEMBERASSERT();
    TWuiObject::Copy(Other, Depth);
    }

AWuiObject  *TWuiContainer::Clone(const TWuiName NewName, int Depth)
    {   MEMBERASSERT();
    ASSERT(Children() != NULL);
    ASSERT(Attributes() != NULL);
    AWuiObject *Result = new TWuiContainer(NewName,
            Attributes()->NObjects(), Children()->NObjects());
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }
TWuiObjects *TWuiContainer::Children() const
    {   MEMBERASSERT();
    return Children_;
    }

int     TWuiContainer::IsA(const TWuiName Class, const TWuiName Method)
    {   MEMBERASSERT();
    if(Class == C_TCONTAINER)
        {
        if(Method == "")
            return TRUE;
        else if(GetAttribute(Method) != NULL)
            return TRUE;
        }
    return FALSE;
    }

AWuiObject *TWuiRoot::ClassConstructor(const TWuiName Name)
    {
    return new TWuiRoot(Name);
    }

// TWuiRoot - Special class for root node.
TWuiRoot::TWuiRoot(const TWuiName ObjectName)
    :   TWuiContainer(S_ROOT, 2)
    {   MEMBERASSERT();

#if defined(__SC__)
    char *ThisCompiler = "Symantec C++ v6.1";
#elif defined(__BORLANDC__)
    char *ThisCompiler = "Borland C++ v4.0";
#elif defined(_MSC_VER) 
    char *ThisCompiler = "Microsoft C++ v8.00c";
#else
    char *ThisCompiler = "<Unknown compiler>";
#endif
    Compiler        = StringClone(ThisCompiler);
    WuimanVersion   = StringClone("0.0");
    DefineAttribute(new TWuiStringProperty(P_COMPILER, &Compiler));
    DefineAttribute(new TWuiStringProperty(P_WUIMANVERSION, &WuimanVersion));
    }

TWuiName    TWuiRoot::ClassName() const
    {   MEMBERASSERT();
    return C_TROOT;
    }
TWuiName    TWuiRoot::ParentClassName() const
    {   MEMBERASSERT();
    return C_TCONTAINER;
    }


// TWuiPath - Handle moving up and down the database "path"
TWuiPath::TWuiPath(char *TotalPath)
    :   AbsolutePath_(TotalPath), RelativePath_(TotalPath)
    {   MEMBERASSERT();
    ASSERT(TotalPath != NULL);
    ASSERT(TotalPath[0] == '/');
    }

// Descend() - Move relative path down one node.
int TWuiPath::Descend()
    {   MEMBERASSERT();
    char    *Child = strchr(RelativePath_, '/');
    if(!Child)   // if no children
        return FALSE;
    // ignore trailers (e.g. "/a/b/c/")
    else if(Child[1] == '\0')
        return FALSE;
    // else there's *some* kind of child name
    else
        {
        RelativePath_   = Child+1;
        return TRUE;
        }
    }

// Ascend() - Move relative path up one node.
int TWuiPath::Ascend()
    {   MEMBERASSERT();
    switch(RelativePath_ - AbsolutePath_)
        {
        case    0:
            return FALSE;
        case    1:
            RelativePath_   = AbsolutePath_;
            return TRUE;
        default:
            RelativePath_   -= 2;
            while(RelativePath_ > AbsolutePath_)
                if(*RelativePath_ == '/')
                    break;
                else
                    --RelativePath_;
            ASSERT(*RelativePath_ == '/');
            ++RelativePath_;
            return TRUE;
        }
    }

// CurrentName() - return first name of relative path.
TWuiName    TWuiPath::CurrentName()
    {   MEMBERASSERT();
    char        Name[TWuiName::MAXNAME+1];
    const char *Rover = strchr(RelativePath(), '/');
    int         Length;
    if(Rover)
        Length  = (int)(Rover-RelativePath());
    else
        Length  = strlen(RelativePath());
    const char *Start = RelativePath();
    // AWuiObject::Get/Set needs this
    if(*Start == '.')
        ++Start;
    strncpy(Name, Start, Length);
    Name[Length]    = '\0';
    return TWuiName(Name);
    }

// Push() - Append to the path.
int     TWuiPath::Push(const char *SubPath)
    {   MEMBERASSERT();
    char   *Rover = AbsolutePath_ + strlen(AbsolutePath_);
    if(Rover > AbsolutePath_)
        --Rover;    // point to char before EOS
    if(*Rover != '/')
        strcat(AbsolutePath_, "/");
    strcat(AbsolutePath_, SubPath);
    return TRUE;    // Need to check for overflow!
    }
// Push() - Append to the path (TWuiName version)
int     TWuiPath::Push(const TWuiName SubPath)
    {   MEMBERASSERT();
    char    Name[TWuiName::MAXNAME+1];
    return Push(SubPath.GetName(Name));
    }

// Pop() - Pop the last node off the path.
int TWuiPath::Pop()
    {   MEMBERASSERT();
    // if already at root, can't pop
    if(!strcmp(AbsolutePath_, "/"))
        return FALSE;
    char   *Rover = AbsolutePath_ + strlen(AbsolutePath_);
    if(Rover > AbsolutePath_)
        --Rover;    // point to char before EOS
    if(Rover > AbsolutePath_ && *Rover == '/')
        --Rover;    // skip any bogus trailing '/'
    while(Rover > AbsolutePath_ && *Rover != '/')
        --Rover;
    // if string did not start with "/", this is bad!
    if(Rover < AbsolutePath_)
        return FALSE;
    else    // else Rover points at '/'
        {
        // as always, special case for root
        if(Rover == AbsolutePath_)
            ++Rover;
        *Rover = '\0';
        return TRUE;
        }
    }

int     TWuiPath::IsLastNode() const
    {   MEMBERASSERT();
    return strchr(RelativePath_, '/') == FALSE;
    }

void    TWuiPath::CopyRelativePath(char *Output) const
    {   MEMBERASSERT();
    strcpy(Output, AbsolutePath_);
    const char *Rover = strchr(RelativePath_, '/');
    if(Rover == NULL)
        Rover   = RelativePath_ + strlen(RelativePath_);
    if(RelativePath_ > AbsolutePath_)
        Output[Rover-AbsolutePath_] = '\0';
    }


////////////////////////////////////////////////////////
// AWuiModifiableAttribute - code common to all attributes.

AWuiModifiableAttribute::AWuiModifiableAttribute(
                const TWuiName ObjectName, int InitialFlags)
    :   AWuiAttribute(ObjectName)
    {   MEMBERASSERT();
    SetFlags(InitialFlags);
    }
void    AWuiModifiableAttribute::BindAddress(void *Address)
    {   MEMBERASSERT();
    if(Address == 0)
        ClearFlags(INDIRECT);
    else
        SetFlags(INDIRECT);
    }

TWuiObjects *AWuiModifiableAttribute::Attributes() const
    {   MEMBERASSERT();
    return PropertyProperties;
    }

long    AWuiModifiableAttribute::Load(TWuiPath Path,
                                          const char *Value)
    {   MEMBERASSERT();
    if(Value != NULL)
        {
        TWuiArgs    Args(Name());
        return Set(Path, &Args, Value);
        }
    else
        return 0;
    }

// Save() - attribute save function.
//          
long    AWuiModifiableAttribute::Save(TWuiPath Path,
                                        int /*Depth*/) const
    {   MEMBERASSERT();
    // if not persistent, we don't save
    if(GetFlags(PERSISTENT) == 0)
        return FALSE;
    long    Result = -1;
    char    SmallBuffer[32];
    // Get value to save by calling derived class function
    // (it will either create string version in buffer we
    //  supply, or return pointer to a string value)
    char *Value = SmallBuffer;
    Result  = AttributeValue(&Value);
    if(Result < WUIMAN_ERRORS)
        DEBUG_FATAL("Can't happen: AttributeValue() failed");
    TCharBuffer Temp(TWuiName::MAXNAME + 16 +strlen(Value));
    char  ThisClass[TWuiName::MAXNAME];
    sprintf(Temp, "%s,%s", ClassName().GetName(ThisClass),
                                                 Value);
    char    AttrName[TWuiName::MAXNAME+1];
    AttrName[0] = '.';
    Name().GetName(AttrName+1);
    DEBUG_TRACE("[AWuiModifiableAttribute::Save] "
        "DB->'%s',attr='%s',val='%s'",
        Path.AbsolutePath(), AttrName, (char *)Temp);
    if(!Database->WriteItem(Path.AbsolutePath(),
                        AttrName, (char *)Temp))
        return WUIMAN_ERR_WRITE_FAILED;
    return TRUE;
    }

long AWuiModifiableAttribute::Set(TWuiPath Path,
                         TWuiArgs *Args, const char *Buffer)
    {   MEMBERASSERT();
    long    Result = -1;
    if(AWuiObject::TraceFlag)
        Trace("AWuiModifiableAttribute::Set", Path, Args);
    if(Args->AttributeName() != Name())
        return AWuiAttribute::Set(Path, Args, Buffer);

    // we should not get called unless "set" was successful
    // Since "set" succeeded, turn off INHERITED bit
    ClearFlags(INHERITED);

    // if PERSISTENT bit is set, save new value to database
    if(GetFlags(PERSISTENT))
        {
        return Save(Path, 1);
#if 0
        char    SmallBuffer[32];
        // Get value to save by calling derived class function
        // (it will either create string version in buffer we
        //  supply, or return pointer to a string value)
        // We could just use "Buffer", but the derived class may
        // have sanitized the string in some way. Also, Buffer
        // will be empty when we are just updating attribute bits
        char *Value = SmallBuffer;
        long ConvertStatus = AttributeValue(&Value);
        if(ConvertStatus < WUIMAN_ERRORS)
            DEBUG_FATAL("Can't happen: AttributeValue() failed");
        if(Result == -1)
            Result  = ConvertStatus;
        TCharBuffer Temp(TWuiName::MAXNAME + 16 +strlen(Value));
        char  ThisClass[TWuiName::MAXNAME];
        sprintf(Temp, "%s,%s", ClassName().GetName(ThisClass),
                                                     Value);
        char    AttrName[TWuiName::MAXNAME+1];
        AttrName[0] = '.';
        Name().GetName(AttrName+1);
        DEBUG_TRACE("[AWuiModifiableAttribute::Set] "
            "DB->'%s',attr='%s',val='%s'",
            Path.AbsolutePath(), AttrName, (char *)Temp);
        if(!Database->WriteItem(Path.AbsolutePath(),
                            AttrName, (char *)Temp))
            return WUIMAN_ERR_WRITE_FAILED;
#endif
        }
    return Result;
    }

long AWuiModifiableAttribute::Get(TWuiPath Path, TWuiArgs *Args,
      char *Buffer, size_t MaxLength)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("AWuiModifiableAttribute::Get", Path, Args);
    // if caller is asking for custom editing
    if(Args->AttributeName() == M_EDITATTR)
        return FALSE;

    // else if we got called as a node under "/Classes"...
    /*else */if(Args->AttributeName() != Name())
        return AWuiObject::Get(Path, Args, Buffer, MaxLength);
    // else this is just a "normal" attribute get value call
    char    SmallBuffer[64];
    SmallBuffer[0]  = '\0';
    // Call derived class to obtain value as string
    char   *Value   = SmallBuffer;
    long    Result  = AttributeValue(&Value);
    if(Buffer != NULL)
        {
        if(Value == NULL)
            Value   = "";
        size_t  ValueLength = strlen(Value) + 1;
        if(ValueLength > MaxLength)
            {
            ValueLength = MaxLength;
            Result      = WUIMAN_ERR_ATTR_TOO_LONG;
            }
        memcpy(Buffer, Value, ValueLength);
        }
    return Result;
    }

///////////////////////////////
// class AWuiProperty

AWuiProperty::AWuiProperty(
                const TWuiName ObjectName, int InitialFlags)
    :   AWuiModifiableAttribute(ObjectName, InitialFlags)
    {   MEMBERASSERT();
    EnforceName('p');
    }

///////////////////////////////
// class TWuiEvent

TWuiEvent::TWuiEvent(
                const TWuiName ObjectName, int InitialFlags)
    :   AWuiModifiableAttribute(ObjectName, InitialFlags)
    {   MEMBERASSERT();
    EnforceName('e');
    }

long    TWuiEvent::AttributeValue(char **Buffer) const
    {   MEMBERASSERT();
    Name().GetName(*Buffer);
    strcat(*Buffer, " (event)");
    return (long)strlen(*Buffer);
    }

///////////////////////////////////////////////////
// class TWuiIntProperty

AWuiObject *TWuiIntProperty::ClassConstructor(const TWuiName Name)
    {
    return new TWuiIntProperty(Name);
    }

TWuiIntProperty::TWuiIntProperty(const TWuiName Name,
                                               int *Address)
    :   AWuiProperty(Name, PERSISTENT|INHERITED|INHERITABLE)
    {   MEMBERASSERT();
    if(Address != NULL)
        {
        IntAddress  = Address;
        SetFlags(INDIRECT);
        }
    else
        IntValue    = 0;
    }

void    TWuiIntProperty::Copy(const AWuiObject *Other1, int Depth)
    {   MEMBERASSERT();
    AWuiProperty::Copy(Other1, Depth);
    const TWuiIntProperty *Other = (TWuiIntProperty *)Other1;
    int *Address;
    if(GetFlags(INDIRECT))
        Address = IntAddress;
    else
        Address = &IntValue;
    const int *OtherAddress;
    if(Other->GetFlags(INDIRECT))
        OtherAddress = Other->IntAddress;
    else
        OtherAddress = &Other->IntValue;
    if(Address && OtherAddress)
        *Address    = *OtherAddress;
    }

AWuiObject *TWuiIntProperty::Clone(const TWuiName NewName, int Depth)
    {   MEMBERASSERT();
    AWuiObject *Result = new TWuiIntProperty(NewName);
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }

long    TWuiIntProperty::AttributeValue(char **Buffer) const
    {   MEMBERASSERT();
    ASSERT(Buffer != NULL);
    ASSERT(*Buffer != NULL);
    const int *Address;
    if(GetFlags(INDIRECT))
        Address = IntAddress;
    else
        Address = &IntValue;
    sprintf(*Buffer, "%d", *Address);
    return *Address;
    }

void    TWuiIntProperty::BindAddress(void *Address)
    {   MEMBERASSERT();
    AWuiModifiableAttribute::BindAddress(Address);
    if(Address != 0)
        IntAddress  = (int *)Address;
    else
        IntValue    = 0;
    }

TWuiName TWuiIntProperty::ClassName() const
    {   MEMBERASSERT();
    return C_TATTRIBUTEINT;
    }
TWuiName TWuiIntProperty::ParentClassName() const
    {   MEMBERASSERT();
    return C_TATTRIBUTE;
    }

long    TWuiIntProperty::Set(TWuiPath Path, TWuiArgs *Args,
                                          const char *Value)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("TWuiIntProperty::Set", Path, Args);
    if(Args && Args->AttributeName() != Name())
        return AWuiModifiableAttribute::Set(Path, Args, Value); 
    char   *End;
    long    Number = strtol(Value, &End, 0);
    if(End == Value || *End != '\0')
        {
        DEBUG_ERROR("Not a legal value for an int"
                " attribute: '%s'", Value);
        return WUIMAN_ERR_SET_BAD_INTEGRAL;
        }
    if(Number > INT_MAX || Number < INT_MIN)
        {
        DEBUG_ERROR("Integer attribute exceeds integer"
                                     " range: '%s'", Value);
        return WUIMAN_ERR_SET_INT_OVERFLOW;
        }
    
    if(GetFlags(INDIRECT))
        *IntAddress = int(Number);
    else
        IntValue    = int(Number);
    return Number;
    }


///////////////////////////////////////////////////
// class TWuiLongProperty

AWuiObject *TWuiLongProperty::ClassConstructor(const TWuiName Name)
    {
    return new TWuiLongProperty(Name);
    }

TWuiLongProperty::TWuiLongProperty(const TWuiName Name,
                                              long *Address)
    :   AWuiProperty(Name, PERSISTENT|INHERITED|INHERITABLE)
    {   MEMBERASSERT();
    if(Address)
        {
        SetFlags(INDIRECT);
        LongAddress = Address;
        }
    else
        LongValue   = 0;
    }

void    TWuiLongProperty::Copy(const AWuiObject *Other1, int Depth)
    {   MEMBERASSERT();
    AWuiProperty::Copy(Other1, Depth);
    const TWuiLongProperty *Other = (TWuiLongProperty *)Other1;
    long *Address;
    if(GetFlags(INDIRECT))
        Address = LongAddress;
    else
        Address = &LongValue;
    const long *OtherAddress;
    if(Other->GetFlags(INDIRECT))
        OtherAddress = Other->LongAddress;
    else
        OtherAddress = &Other->LongValue;
    if(Address && OtherAddress)
        *Address    = *OtherAddress;
    }

AWuiObject *TWuiLongProperty::Clone(const TWuiName NewName, int Depth)
    {   MEMBERASSERT();
    AWuiObject *Result = new TWuiLongProperty(NewName);
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }

long    TWuiLongProperty::AttributeValue(char **Buffer) const
    {   MEMBERASSERT();
    ASSERT(Buffer != NULL);
    ASSERT(*Buffer != NULL);
    const long *Address;
    if(GetFlags(INDIRECT))
        Address = LongAddress;
    else
        Address = &LongValue;
    sprintf(*Buffer, "%ld", *Address);
    return *Address;
    }

void    TWuiLongProperty::BindAddress(void *Address)
    {   MEMBERASSERT();
    AWuiProperty::BindAddress(Address);
    if(Address != 0)
        LongAddress  = (long *)Address;
    else
        LongValue    = 0;
    }

TWuiName TWuiLongProperty::ClassName() const
    {   MEMBERASSERT();
    return C_TATTRIBUTELONG;
    }
TWuiName TWuiLongProperty::ParentClassName() const
    {   MEMBERASSERT();
    return C_TATTRIBUTE;
    }

long    TWuiLongProperty::Set(TWuiPath Path, TWuiArgs *Args,
                                          const char *Value)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("TWuiLongAttribute::Set", Path, Args);
    if(Args && Args->AttributeName() != Name())
        return AWuiProperty::Set(Path, Args, Value); 
    char   *End;
    long    Number = strtol(Value, &End, 0);
    if(End == Value || *End != '\0')
        {
        DEBUG_ERROR("Not a legal value for an long"
                " attribute: '%s'", Value);
        return WUIMAN_ERR_SET_BAD_INTEGRAL;
        }
    
    if(GetFlags(INDIRECT))
        *LongAddress = Number;
    else
        LongValue    = Number;
    return TRUE;
    }

#ifdef WUI_FLOAT

///////////////////////////////////////////////////
// class TWuiDoubleProperty


AWuiObject *TWuiDoubleProperty::ClassConstructor(const TWuiName Name)
    {
    return new TWuiDoubleProperty(Name);
    }

TWuiDoubleProperty::TWuiDoubleProperty(const TWuiName Name,
                                              double *Address)
    :   AWuiProperty(Name, PERSISTENT|INHERITED|INHERITABLE)
    {   MEMBERASSERT();
    if(Address)
        {
        SetFlags(INDIRECT);
        DoubleAddress = Address;
        }
    else
        DoubleValue   = 0;
    }

void    TWuiDoubleProperty::Copy(const AWuiObject *Other1, int Depth)
    {   MEMBERASSERT();
    AWuiProperty::Copy(Other1, Depth);
    const TWuiDoubleProperty *Other = (TWuiDoubleProperty *)Other1;
    double *Address;
    if(GetFlags(INDIRECT))
        Address = DoubleAddress;
    else
        Address = &DoubleValue;
    const double *OtherAddress;
    if(Other->GetFlags(INDIRECT))
        OtherAddress = Other->DoubleAddress;
    else
        OtherAddress = &Other->DoubleValue;
    if(Address && OtherAddress)
        *Address    = *OtherAddress;
    }

AWuiObject *TWuiDoubleProperty::Clone(const TWuiName NewName, int Depth)
    {   MEMBERASSERT();
    AWuiObject *Result = new TWuiDoubleProperty(NewName);
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }

long    TWuiDoubleProperty::AttributeValue(char **Buffer) const
    {   MEMBERASSERT();
    ASSERT(Buffer != NULL);
    ASSERT(*Buffer != NULL);
    const double *Address;
    if(GetFlags(INDIRECT))
        Address = DoubleAddress;
    else
        Address = &DoubleValue;
    sprintf(*Buffer, "%g", *Address);
    return *Address;
    }

void    TWuiDoubleProperty::BindAddress(void *Address)
    {   MEMBERASSERT();
    AWuiProperty::BindAddress(Address);
    if(Address != 0)
        DoubleAddress  = (double *)Address;
    else
        DoubleValue    = 0;
    }

TWuiName TWuiDoubleProperty::ClassName() const
    {   MEMBERASSERT();
    return C_TATTRIBUTEDOUBLE;
    }
TWuiName TWuiDoubleProperty::ParentClassName() const
    {   MEMBERASSERT();
    return C_TATTRIBUTE;
    }

long    TWuiDoubleProperty::Set(TWuiPath Path, TWuiArgs *Args,
                                          const char *Value)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("TWuiDoubleProperty::Set", Path, Args);
    if(Args && Args->AttributeName() != Name())
        return AWuiProperty::Set(Path, Args, Value); 
    char   *End;
    double    Number = strtod(Value, &End);
    if(End == Value || *End != '\0')
        {
        DEBUG_ERROR("Not a legal value for a double"
                " attribute: '%s'", Value);
        return WUIMAN_ERR_SET_BAD_INTEGRAL;
        }
    
    if(GetFlags(INDIRECT))
        *DoubleAddress = Number;
    else
        DoubleValue    = Number;
    return TRUE;
    }

#endif  // WUI_FLOAT


///////////////////////////////////////////////////
// String attributes

AWuiObject *TWuiStringProperty::ClassConstructor(const TWuiName Name)
    {
    return new TWuiStringProperty(Name);
    }

TWuiStringProperty::TWuiStringProperty(const TWuiName Name,
                                             char **Address)
    :   AWuiProperty(Name)
    {   MEMBERASSERT();
    if(Address != NULL)
        {
        SetFlags(INDIRECT);
        StringAddress   = Address;
        }
    else
        StringValue = 0;
    }


TWuiStringProperty::~TWuiStringProperty()
    {   MEMBERASSERT();
    char    **Address = &StringValue;
    if(GetFlags(INDIRECT))
        Address = StringAddress;
    if(*Address != NULL)
        {
        delete[] *Address;
        *Address    = NULL;
        }
    }
void    TWuiStringProperty::Copy(const AWuiObject *Other1, int Depth)
    {   MEMBERASSERT();
    AWuiProperty::Copy(Other1, Depth);
    const TWuiStringProperty *Other = (TWuiStringProperty *)Other1;
    char   **Address;
    if(GetFlags(INDIRECT))
        Address = StringAddress;
    else
        Address = &StringValue;
    char *const*OtherAddress;
    if(Other->GetFlags(INDIRECT))
        OtherAddress = Other->StringAddress;
    else
        OtherAddress = &Other->StringValue;
    if(Address && OtherAddress)
        {
        if(*Address != NULL)
            {
            delete[] *Address;
            *Address = 0;
            }
        if(*OtherAddress)
            *Address    = StringClone(*OtherAddress);
        else
            *Address    = NULL;
        }
    }
AWuiObject *TWuiStringProperty::Clone(const TWuiName Name, int Depth)
    {   MEMBERASSERT();
    AWuiObject *Result = new TWuiStringProperty(Name);
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }
void    TWuiStringProperty::BindAddress(void *Address)
    {   MEMBERASSERT();
DEBUG_TRACE("!!!!!!!!!!!!!!BindAddress called!" );
    if(!GetFlags(INDIRECT))
        {
        if(StringValue != 0)
            {
            delete[] StringValue;
            StringValue = 0;
            }
        }
    AWuiProperty::BindAddress(Address);
    if(Address != 0)
        StringAddress  = (char **)Address;
    else
        StringValue    = 0;
    }


long    TWuiStringProperty::AttributeValue(char **Buffer) const
    {   MEMBERASSERT();
    ASSERT(Buffer != NULL);
    ASSERT(*Buffer != NULL);
    char **Address;
    if(GetFlags(INDIRECT))
        Address = (char **)StringAddress;
    else
        Address = (char **)&StringValue;
    ASSERT(Address != NULL);
DEBUG_TRACE("String AV = '%s', indirect = %d", *Address, GetFlags(INDIRECT));
    *Buffer = *Address;
    return TRUE;
    }

TWuiName TWuiStringProperty::ClassName() const
    {   MEMBERASSERT();
    return C_TATTRIBUTESTRING;
    }

TWuiName TWuiStringProperty::ParentClassName() const
    {   MEMBERASSERT();
    return C_TATTRIBUTE;
    }

int     TWuiStringProperty::IsA(const TWuiName Class, const TWuiName Method)
    {   MEMBERASSERT();
    if(Class == C_TATTRIBUTESTRING)
        {
        if(Method == "")
            return TRUE;
        else if(GetAttribute(Method) != NULL)
            return TRUE;
        else
            return FALSE;
        }
    else
        return AWuiAttribute::IsA(Class, Method);
    }



// Set - delete our current value, make a copy of the new string.
long    TWuiStringProperty::Set(TWuiPath Path,
                                        TWuiArgs *Args,
                                         const char *Buffer)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("TWuiStringProperty::Set", Path, Args);
    // if it's us, then handle actual set value
    if(Args->AttributeName() == Name())
        {
        char    **Address = GetFlags(INDIRECT)?StringAddress:&StringValue;
        ASSERT(Address != NULL);
        if(*Address)
            {
            delete[] *Address;
            *Address    = 0;
            }
        *Address    = Buffer ? StringClone(Buffer) : 0;
        }
    return AWuiModifiableAttribute::Set(Path, Args, Buffer); 

    }


TWuiMethod::TWuiMethod(const TWuiName ObjectName)
    :   AWuiAttribute(ObjectName)
    {   MEMBERASSERT();
    EnforceName('m');
    }

void    TWuiMethod::Copy(const AWuiObject *Other, int Depth)
    {   MEMBERASSERT();
    AWuiAttribute::Copy(Other, Depth);
    }

AWuiObject *TWuiMethod::Clone(const TWuiName ObjectName, int Depth)
    {   MEMBERASSERT();
    AWuiObject *Result = new TWuiMethod(ObjectName);
    if(Result == NULL)
        return Result;
    else
        {
        Result->Copy(this, Depth);
        return Result;
        }
    }

long TWuiMethod::AttributeValue(char **Buffer) const
    {   MEMBERASSERT();
    ASSERT(Buffer != NULL);
    ASSERT(*Buffer != NULL);
    **Buffer    = '\0';
    return 0;
    }

TWuiName    TWuiMethod::ClassName() const
    {   MEMBERASSERT();
    return C_TATTRIBUTEMETHOD;
    }
TWuiName    TWuiMethod::ParentClassName() const
    {   MEMBERASSERT();
    return C_TATTRIBUTE;
    }

AWuiAttribute::AWuiAttribute(const TWuiName ObjectName)
    :   AWuiObject(ObjectName)
    {   MEMBERASSERT();
    }

void    AWuiAttribute::EnforceName(char Prefix)
    {   MEMBERASSERT();
    char    ThisName[TWuiName::MAXNAME+1];
    Name().GetName(ThisName);
    if(ThisName[0] != Prefix || ThisName[1] != '_')
        {
        memmove(ThisName+2, ThisName, TWuiName::MAXNAME-2);
        ThisName[0] = Prefix;
        ThisName[1] = '_';
        ThisName[TWuiName::MAXNAME] = '\0';
        Rename(ThisName);
        }
    }

TWuiName    AWuiAttribute::ClassName() const
    {   MEMBERASSERT();
    return C_TATTRIBUTE;
    }

TWuiName    AWuiAttribute::ParentClassName() const
    {   MEMBERASSERT();
    return 0;
    }

int     AWuiAttribute::IsA(const TWuiName Class, const TWuiName Method)
    {   MEMBERASSERT();
    if(Class == C_TATTRIBUTE)
        {
        if(Method == "")
            return TRUE;
        else if(GetAttribute(Method) != NULL)
            return TRUE;
        }
    // we are top-most WUI class, so no parent to resort to
    return FALSE;
    }


///////////////////////////////////////////////////
// Internal Bit attributes

TWuiIntPropertyBit::TWuiIntPropertyBit(
            const TWuiName Name)
    :   AWuiProperty(Name)
    {   MEMBERASSERT();
    SetFlags(REFERENCE|PERSISTENT);
    }

void    TWuiIntPropertyBit::Copy(const AWuiObject * /*Other*/, int /*Depth*/)
    {   MEMBERASSERT();
    // do nothing, copying of bits is handled by
    // AWuiObject::Copy()
    return;
    }
AWuiObject *TWuiIntPropertyBit::Clone(
        const TWuiName /*Name*/, int /*Depth*/)
    {   MEMBERASSERT();
    ASSERT(FALSE);  // Should never be called!
    return NULL;
    }


TWuiName TWuiIntPropertyBit::ClassName() const
    {   MEMBERASSERT();
    return C_TATTRIBUTEINTERNALBIT;
    }
TWuiName TWuiIntPropertyBit::ParentClassName() const
    {   MEMBERASSERT();
    return C_TATTRIBUTEINT;
    }
long    TWuiIntPropertyBit::AttributeValue(char **Buffer) const
    {   MEMBERASSERT();
    ASSERT(Buffer != NULL);
    ASSERT(*Buffer != NULL);

    int On = GetFlags(0x8000);
    if(On != 0)
        strcpy(*Buffer, "1");
    else
        strcpy(*Buffer, "0");
    return On != 0;
    }

long    TWuiIntPropertyBit::Set(TWuiPath Path, TWuiArgs *Args,
                                          const char *Value)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("TWuiIntPropertyBit::Set", Path, Args);
    int On   = 0;
    if(Value != NULL)
        On  = (atoi(Value) != 0);
    // very tacky hack. essentially reserving topmost flag bit
    // to communicate with our AttributeValue() function across
    // the call to AWuiModifiableAttribute::Set().  Ugh.
    SetFlags(0x8000, On);
    return AWuiProperty::Set(Path, Args, Value);
    }

